import numpy as np
from airsim import MultirotorClient
from envs import BaseEnvironment
import abc
import math
from msgpackrpc.error import TimeoutError

from numpy.typing import NDArray

from tf_agents.environments.py_environment import PyEnvironment
from tf_agents.trajectories import time_step as ts
from tf_agents.typing.types import TimeStep, NestedArray, Text

class BaseEnvironment(PyEnvironment):
    min_throttle = 0.4
    max_throttle = 1.0

    max_pitch = math.pi / 4
    min_pitch = -max_pitch

    max_roll = math.pi / 4
    min_roll = -max_roll

    max_yaw = math.pi / 2
    min_yaw = -max_yaw
    def __init__(self, ip: str = None, client: MultirotorClient = None, handle_auto_reset: bool = False):
        self._last_inputs = None
        self._ip = ip
        if client is None:
            self._create_client()
        else:
            self._client = client

        super(BaseEnvironment, self).__init__(handle_auto_reset)

    def _reset(self) -> TimeStep:
        self._last_inputs = None
        self._client.reset()
        self._client.confirmConnection()
        self._client.cancelLastTask()
        self._client.enableApiControl(True)
        self._client.armDisarm(True)
        return ts.restart(self._get_observation())
    
    def _step(self, action: NestedArray) -> TimeStep:
        """
        action[0]: vz
        action[1]: vx
        action[2]: vy
        """
        if self._must_restart():
            return self.reset()

        throttle = self._convert_throttle_for_airsim(action[0].item())
        pitch = self._convert_pitch_for_airsim(action[1].item())
        roll = self._convert_roll_for_airsim(action[2].item())
        yaw = self._convert_yaw_for_airsim(action[3].item())
        current_inputs = (throttle, pitch, roll, yaw)

        self._client.moveByRollPitchYawrateThrottleAsync(roll, pitch, yaw, throttle, 0.01).join()

        obs = self._get_observation()
        reward = self._calculate_reward(obs, current_inputs, self._last_inputs)

        if self._is_terminated(obs):
            return ts.termination(obs, reward)

        if self._is_truncated(obs):
            return ts.truncation(obs, reward)
        
        self._last_inputs = current_inputs         
        return ts.transition(obs, reward=reward)
    
    def _create_client(self):
        self._client = MultirotorClient(ip=self._ip, timeout_value=10)
    
    @staticmethod
    def _convert_property_for_airsim(min: float, max: float, property: np.float32) -> float:
        return (max - min) * property + min
    
    def _convert_throttle_for_airsim(self, throttle: np.float32) -> float:
        return BaseEnvironment._convert_property_for_airsim(self.min_throttle, self.max_throttle, throttle)
    
    def _convert_pitch_for_airsim(self, pitch: np.float32) -> float:
        return BaseEnvironment._convert_property_for_airsim(self.min_pitch, self.max_pitch, pitch)
    
    def _convert_roll_for_airsim(self, roll: np.float32) -> float:
        return BaseEnvironment._convert_property_for_airsim(self.min_roll, self.max_roll, roll)
    
    def _convert_yaw_for_airsim(self, yaw: np.float32) -> float:
        return BaseEnvironment._convert_property_for_airsim(self.min_yaw, self.max_yaw, yaw)
    
    def render(self, mode: Text = 'rgb_array') -> NestedArray | None:
        raise NotImplementedError()
    
    @abc.abstractmethod
    def _is_terminated(self, obs: NDArray) -> bool:
        """"""

    @abc.abstractmethod
    def _is_truncated(self, obs: NDArray) -> bool:
        """"""

    @abc.abstractmethod
    def _must_restart(self) -> bool:
        """"""
    
    @abc.abstractmethod
    def _get_observation(self) -> NDArray:
        """"""

    @abc.abstractmethod
    def _calculate_reward(self, obs: NDArray, current_inputs: NDArray, last_inputs: NDArray) -> np.float64:
        """"""
