from airsim import MultirotorClient
from numpy.typing import NDArray
from tf_agents.trajectories import time_step as ts
from tf_agents.typing import types
from envs import BaseEnvironment, NavigationEnvironment
import numpy as np
import math
import airsim

from tf_agents.specs.array_spec import BoundedArraySpec, ArraySpec
from tf_agents.typing.types import TimeStep, NestedArray, NestedArraySpec

class NavigationEnvironment(BaseEnvironment):
    latitude_longitude_amplifier = 10000
    min_pitch = 0
    def __init__(self,
                 ip: str = None,
                 client: MultirotorClient = None,
                 handle_auto_reset: bool = False,
                 dynamic_end: bool = False,
                 initial_height: np.float64 = 550.0,
                 desired_latitude: np.float64 = 47.63748371181762,
                 desired_longitude: np.float64 = -122.1442011793522,
                 desired_height: np.float64 = 550.0):
        self._target_latitude = desired_latitude
        self._target_longitude = desired_longitude
        self._target_height = desired_height
        self._initial_height = initial_height
        self._needs_restart = False
        self._dynamic_end = dynamic_end
        self._last_steps = []

        self._action_spec = BoundedArraySpec(shape=(3,), dtype=np.float32, minimum=0, maximum=1, name='action')
        self._observation_spec = ArraySpec(shape=(7,), dtype=np.float64, name='observation')

        super(NavigationEnvironment, self).__init__(ip, client, handle_auto_reset)

    def action_spec(self) -> BoundedArraySpec:
        return self._action_spec
    
    def observation_spec(self) -> ArraySpec:
        return self._observation_spec
    
    def _reset(self) -> TimeStep:
        super()._reset()
        self._needs_restart = False
        if self._dynamic_end:
            self._target_latitude = np.random.random_sample() * 0.006238241 + 47.63496827504901
            self._target_longitude = np.random.random_sample() * 0.025975688 - 122.15618176959885
            self._target_height = self._initial_height
        
        position = self._client.getMultirotorState().gps_location
        self._client.moveToGPSAsync(latitude=position.latitude, longitude=position.longitude, altitude=self._initial_height, velocity=20).join()
        return ts.restart(self._get_observation())
    
    def _step(self, action: NestedArray) -> TimeStep:
        full_action = np.insert(action, 2, [0.5])
        return super()._step(full_action)

    def _is_terminated(self, obs: NDArray) -> bool:
        x_delta = obs[0]
        y_delta = obs[1]
        altitude_delta = obs[2]

        path_distance = math.sqrt(x_delta ** 2 + y_delta ** 2 + altitude_delta ** 2)
        return path_distance < 0.001
    
    def _is_truncated(self, obs: NDArray) -> bool:
        return False
    
    def _must_restart(self) -> bool:
        return self._needs_restart
    
    def _get_observation(self) -> NDArray:
        state = self._client.getMultirotorState()
        geo_point = state.gps_location
        altitude_delta = self._target_height - geo_point.altitude
        latitude_delta = self._target_latitude - geo_point.latitude
        longitude_delta = self._target_longitude - geo_point.longitude

        altitude_delta_km = self._meters_to_km(altitude_delta)
        x_delta, y_delta = self._latlon_to_xy(latitude_delta, longitude_delta)

        velocity = state.kinematics_estimated.linear_velocity

        normalized_latitude = latitude_delta / math.sqrt(latitude_delta ** 2 + longitude_delta ** 2)
        normalized_longitude = longitude_delta / math.sqrt(latitude_delta ** 2 + longitude_delta ** 2)

        orientation_angles = airsim.to_eularian_angles(state.kinematics_estimated.orientation)
        heading_delta = math.atan2(normalized_longitude, normalized_latitude) - orientation_angles[2]
        heading_delta = self._normalize_rad(heading_delta)

        return np.array([x_delta, y_delta, altitude_delta_km, heading_delta, velocity.x_val, velocity.y_val, velocity.z_val], dtype=np.float64)

    def _lat_long_to_meters(self, lat_long: float) -> float:
        return lat_long * 111139
    
    def _meters_to_km(self, meters: float) -> float:
        return meters / 1000
    
    def _latlon_to_xy(self, latitude, longitude):
        # Radius of the Earth (in kilometers)
        R = 6371.0

        # Convert latitude and longitude from degrees to radians
        lat_rad = math.radians(latitude)
        lon_rad = math.radians(longitude)

        # Calculate x and y using a simple cylindrical projection
        x = R * lon_rad
        y = R * lat_rad

        return x, y
    
    def _normalize_rad(self, degree: float) -> float:
        while degree < -1 * math.pi:
            degree += 2 * math.pi

        while degree > math.pi:
            degree -= 2* math.pi

        return degree
    
    def _calculate_reward(self, obs: NDArray, current_inputs: NDArray, last_inputs: NDArray) -> np.float64:
        x_delta = obs[0]
        y_delta = obs[1]
        altitude_delta = obs[2]
        heading_delta = obs[3]

        path_penalty = -math.sqrt(x_delta ** 2 + y_delta ** 2 + altitude_delta ** 2)
        x_y_delta = math.sqrt(x_delta ** 2 + y_delta ** 2)
        heading_penalty = -abs(heading_delta) if x_y_delta > 0.001 else 0 # prevent big heading changes when target is reached
        altitude_penalty = -abs(altitude_delta)

        inputs_penalty = 0 # not needed currently, but would make drone inputs smoother
        if last_inputs is not None:
            pitch_penalty = -abs(current_inputs[1] - last_inputs[1])
            yaw_penalty = -abs(current_inputs[3] - last_inputs[3])

            inputs_penalty = pitch_penalty + yaw_penalty

        cummulative_reward = path_penalty + heading_penalty

        print("Path reward: {:.5f}, heading reward: {:.5f}, cummulative reward: {:.5f}, x delta: {:.5f}, y delta: {:.5f}, z delta {:.5f}".format(path_penalty, heading_penalty, cummulative_reward, x_delta, y_delta, altitude_delta))
        
        return np.float64(cummulative_reward)
