from airsim import MultirotorClient
from numpy.typing import NDArray
from envs import BaseEnvironment, StartEnvironment
import numpy as np
from tf_agents.trajectories import time_step as ts

from tf_agents.specs.array_spec import BoundedArraySpec, ArraySpec
from tf_agents.typing.types import TimeStep, NestedArray

class StartEnvironment(BaseEnvironment):
    def __init__(self, 
                 ip: str = None, 
                 client: MultirotorClient = None, 
                 handle_auto_reset: bool = False, 
                 desired_height: np.float64=140.0, 
                 check_out_of_bounds: bool = False,
                 dynamic_start: bool = False):
        self._desired_height = desired_height
        self._check_out_of_bounds = check_out_of_bounds
        self._is_out_of_bounds = False
        self._dynamic_start = dynamic_start
        self._last_steps = []

        self._action_spec = BoundedArraySpec(shape=(1,), dtype=np.float32, minimum=0, maximum=1, name='action')
        self._observation_spec = ArraySpec(shape=(2,), dtype=np.float64, name='observation')

        super(StartEnvironment, self).__init__(ip, client, handle_auto_reset)
    
    def action_spec(self) -> BoundedArraySpec:
        return self._action_spec
    
    def observation_spec(self) -> ArraySpec:
        return self._observation_spec
    
    def _reset(self) -> TimeStep:
        self._is_out_of_bounds = False
        time_step = super()._reset()
        if self._dynamic_start:
            altitude = np.random.random_sample() * 60 + 122.2
            position = self._client.getMultirotorState().gps_location
            self._client.moveToGPSAsync(latitude=position.latitude, longitude=position.longitude, altitude=altitude, velocity=10).join()
            time_step = ts.restart(self._get_observation())
        return time_step
    
    def _step(self, action: NestedArray) -> TimeStep:
        full_action = np.append(action, [0.5, 0.5, 0.5])
        ts = super()._step(full_action)
        print("Current height delta: {}".format(ts.observation[0]))

        if len(self._last_steps) >= 50:
            self._last_steps.pop()
        self._last_steps.append(ts)
        return ts

    def _is_terminated(self, obs: NDArray) -> bool:
        return False
    
    def _is_truncated(self, obs: NDArray) -> bool:
        self._is_out_of_bounds = self._check_out_of_bounds and obs[0] > 30
        return self._is_out_of_bounds
    
    def _must_restart(self) -> bool:
        return self._is_out_of_bounds
    
    def _get_observation(self) -> NDArray:
        state = self._client.getMultirotorState()
        geo_point = state.gps_location
        height_delta = geo_point.altitude - self._desired_height
        velocity = state.kinematics_estimated.linear_velocity
        return np.array([height_delta, velocity.z_val], dtype=np.float64)
    
    def _calculate_reward(self, obs: NDArray, current_inputs: NDArray, last_inputs: NDArray) -> np.float32:
        return np.float32(-abs(obs[0]))
