from typing import Any, Tuple
import tensorflow as tf
import reverb
import os
import numpy as np

from tf_agents.typing.types import Float, Int, NestedTensorSpec
from tf_agents.agents.ddpg.critic_network import CriticNetwork
from tf_agents.networks.actor_distribution_network import ActorDistributionNetwork
from tf_agents.agents.sac.tanh_normal_projection_network import TanhNormalProjectionNetwork
from tf_agents.agents.sac.sac_agent import SacAgent
from tf_agents.replay_buffers.reverb_replay_buffer import ReverbReplayBuffer
from tf_agents.replay_buffers.reverb_utils import ReverbAddTrajectoryObserver
from tf_agents.policies.tf_policy import TFPolicy
from tf_agents.agents.tf_agent import TFAgent
from tf_agents.policies.py_tf_eager_policy import PyTFEagerPolicy
from tf_agents.train.actor import Actor
from tf_agents.environments.py_environment import PyEnvironment
from tf_agents.environments.tf_py_environment import TFPyEnvironment
from tf_agents.policies.py_policy import PyPolicy
from tf_agents.metrics.py_metrics import EnvironmentSteps
from tf_agents.train import actor
from tf_agents.train import learner
from tf_agents.train.learner import Learner
from tf_agents.utils.common import Checkpointer
from tf_agents.policies import PolicySaver

def set_gpu():
    gpus = tf.config.list_physical_devices('GPU')
    for gpu in gpus:
        tf.config.experimental.set_memory_growth(gpu, True)

def create_agent(action_spec: Any, 
                 observation_spec: Any,
                 time_step_spec: Any,
                 train_step: tf.Variable,
                 actor_fc_layer_params: Tuple,
                 critic_joint_fc_layer_params: Tuple,
                 actor_learning_rate: Float = 3e-4,
                 critic_learning_rate: Float = 3e-4,
                 alpha_learning_rate: Float = 3e-4,
                 target_update_tau: Float = 0.005,
                 target_update_period: Int = 1,
                 gamma: Float = 0.99,
                 reward_scale_factor: Float = 1.0,
                 ) -> SacAgent:
    critic_net = CriticNetwork(
        (observation_spec, action_spec),
        observation_fc_layer_params=None,
        action_fc_layer_params=None,
        joint_fc_layer_params=critic_joint_fc_layer_params,
        kernel_initializer='glorot_uniform',
        last_kernel_initializer='glorot_uniform')

    actor_net = ActorDistributionNetwork(
        observation_spec,
        action_spec,
        fc_layer_params=actor_fc_layer_params,
        continuous_projection_net=(
            TanhNormalProjectionNetwork))
    
    return SacAgent(
        time_step_spec,
        action_spec,
        actor_network=actor_net,
        critic_network=critic_net,
        actor_optimizer=tf.keras.optimizers.Adam(learning_rate=actor_learning_rate),
        critic_optimizer=tf.keras.optimizers.Adam(learning_rate=critic_learning_rate),
        alpha_optimizer=tf.keras.optimizers.Adam(learning_rate=alpha_learning_rate),
        target_update_tau=target_update_tau,
        target_update_period=target_update_period,
        td_errors_loss_fn=tf.math.squared_difference,
        gamma=gamma,
        reward_scale_factor=reward_scale_factor,
        train_step_counter=train_step)

def create_replay_buffer(collect_data_spec: NestedTensorSpec, replay_buffer_capacity: int) -> Tuple[ReverbReplayBuffer, ReverbAddTrajectoryObserver]:
    table_name = 'uniform_table'
    table = reverb.Table(
        table_name,
        max_size=replay_buffer_capacity,
        sampler=reverb.selectors.Uniform(),
        remover=reverb.selectors.Fifo(),
        rate_limiter=reverb.rate_limiters.MinSize(1))

    reverb_server = reverb.Server([table])
    replay_buffer = ReverbReplayBuffer(
        collect_data_spec,
        sequence_length=2,
        table_name=table_name,
        local_server=reverb_server)
    
    rb_observer = ReverbAddTrajectoryObserver(
        replay_buffer.py_client,
        table_name,
        sequence_length=2,
        stride_length=1)
    
    return replay_buffer, rb_observer

def create_policy(policy: TFPolicy) -> PyTFEagerPolicy:
    return PyTFEagerPolicy(policy, use_tf_function=True)

def create_actors(train_env: PyEnvironment, 
                  eval_env: PyEnvironment, 
                  random_policy: PyPolicy, 
                  collect_policy: PyPolicy, 
                  eval_policy: PyPolicy,
                  train_step: tf.Variable,
                  rb_observer: ReverbAddTrajectoryObserver,
                  tempdir: str = "out/"
                  ) -> Tuple[Actor, Actor, Actor]:
    random_actor = Actor(
        train_env,
        random_policy,
        train_step,
        steps_per_run=1,
        observers=[rb_observer])

    env_step_metric = EnvironmentSteps()
    collect_actor = Actor(
        train_env,
        collect_policy,
        train_step,
        steps_per_run=1,
        metrics=actor.collect_metrics(10),
        summary_dir=os.path.join(tempdir, learner.TRAIN_DIR),
        observers=[rb_observer, env_step_metric])

    eval_actor = Actor(
        eval_env,
        eval_policy,
        train_step,
        steps_per_run=1,
        metrics=actor.eval_metrics(1),
    )

    return random_actor, collect_actor, eval_actor

def create_learner(agent: TFAgent, train_step: tf.Variable, tempdir: str, replay_buffer: ReverbReplayBuffer, batch_size: int) -> Learner:
    dataset = replay_buffer.as_dataset(
        sample_batch_size=batch_size, num_steps=2).prefetch(50)
    experience_dataset_fn = lambda: dataset

    return Learner(
    tempdir,
    train_step,
    agent,
    experience_dataset_fn)

def create_checkpoint(tempdir: str, agent: TFAgent, replay_buffer: ReverbReplayBuffer, train_step: tf.Variable) -> Checkpointer:
    checkpoint_dir = os.path.join(tempdir, 'checkpoint')
    checkpointer = Checkpointer(
        ckpt_dir=checkpoint_dir,
        max_to_keep=10,
        agent=agent,
        policy=agent.policy,
        replay_buffer=replay_buffer,
        global_step=train_step
    )
    checkpointer.initialize_or_restore()

    policy_saver = PolicySaver(agent.policy)

    return checkpointer, policy_saver

def train_loop(learner: Learner, 
               collect_actor: Actor, 
               eval_actor: Actor,
               random_actor: Actor,
               train_env: TFPyEnvironment, 
               eval_env: TFPyEnvironment, 
               checkpointer: Checkpointer,
               policy_saver: PolicySaver,
               train_step: tf.Variable, 
               train_episodes: int, 
               train_steps_per_episode: int,
               initial_steps: int = 1000,
               initial_learning_iterations: int = 1,
               eval_interval: int = 5,
               eval_steps: int = 1000,
               is_initial: bool = True,
               outdir: str = "out",
               preceding_train_steps: int = 0
               ):
    policy_dir = os.path.join(outdir, "policy")
    logs_dir = os.path.join(outdir, "eval")
    train_summary_writer = tf.summary.create_file_writer(logs_dir)

    for _ in range(initial_steps):
        if is_initial:
            random_actor.run()
        else:
            collect_actor.run()

    learner.run(iterations=initial_learning_iterations)
    for i in range(train_episodes):
        train_env.reset()

        for _ in range(train_steps_per_episode):
            collect_actor.run()

            learner.run(iterations=1)
        checkpointer.save(train_step)
        policy_saver.save(policy_dir)

        if i % eval_interval == 0:
            print("######## Evaluation episode {} ########".format(i))
            ts = eval_env.reset()

            cummulative_reward = ts.reward.numpy().item()
            for _ in range(eval_steps): # since some envs don't have a terminal state, this must be done manually instead of using the actor
                action = eval_actor.policy.action(ts)
                ts = eval_env.step(action)

                cummulative_reward += ts.reward.numpy().item()

                if ts.is_last():
                    break

            with train_summary_writer.as_default():
                tf.summary.scalar(name='CummulativeReward', data=cummulative_reward, step=i * train_steps_per_episode + preceding_train_steps)

            train_summary_writer.flush()            

def is_initial(tempdir: str) -> bool:
    return not os.path.exists(tempdir)
